from functools import wraps
from typing import Union, Tuple

from bools.log import Logger
from bools.datetime import Datetime


def catch(_func=None, *, exception: Union[Tuple[type(Exception), ...], type(Exception)] = Exception,
          except_func=None, except_return=None, log='', print_traceback=True):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exception:
                import traceback
                if except_func is not None:
                    except_func()
                if log:
                    Logger.error(log)
                if print_traceback:
                    print(traceback.print_exc())
                if except_return is not None:
                    return except_return

        return wrapper

    return decorator(_func) if _func else decorator


def timeit(_func=None, *, count=3, return_costs=False):
    def decorator(func):
        @wraps(func)
        def wrapper(*arg, **kwargs):
            current, seconds = Datetime.now().timestamp(), []
            for i in range(count):
                func(*arg, **kwargs)
                now = Datetime.now().timestamp()
                seconds.append(now - current)
                current = now
            Logger.info(f'平均执行时间: {sum(seconds) / count:.3f}s')
            if return_costs:
                return seconds

        return wrapper

    return decorator(_func) if _func else decorator


def parallel(func=None, count: int = None):
    setattr(_outer_func, _FUNC, func)

    def inner(data: list):
        from multiprocessing import Pool
        with Pool(count) as p:
            result = p.map(_outer_func, data)
        return result

    return inner


_FUNC = 'func'


def _outer_func(*args, **kwargs):
    func = getattr(_outer_func, _FUNC)
    return func(*args, **kwargs) if func.__code__.co_argcount else func()



import logging
import colorlog
from functools import wraps


def _check(func):
    @wraps(func)
    def decorator(*args, **kwargs):
        if Logger.logger is None:
            Logger.init()
        return func(*args, **kwargs)

    return decorator


class Logger:
    logger = None

    @classmethod
    def init(cls, log_level=logging.INFO):
        handler = colorlog.StreamHandler()
        formatter = colorlog.ColoredFormatter(
            "%(log_color)s[%(asctime)s][%(levelname)s] : %(message)s",
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red'
            },
            reset=True
        )

        handler.setFormatter(formatter)
        logger = colorlog.getLogger('root')
        logger.addHandler(handler)
        logger.setLevel(log_level)
        cls.logger = logger

    @classmethod
    @_check
    def debug(cls, msg):
        cls.logger.debug(msg)

    @classmethod
    @_check
    def info(cls, msg):
        cls.logger.info(msg)

    @classmethod
    @_check
    def warning(cls, msg):
        cls.logger.warning(msg)

    @classmethod
    @_check
    def error(cls, msg):
        cls.logger.error(msg)



import requests
import json
import re
from dataclasses import dataclass
from typing import Iterator, Generator
from itertools import islice

from .dbc import DBC, http_json_res_parse
from bools.functools import catch
from bools.log import Logger

TEMPLATE_NAME = 'bowaer'
_HEADERS = {
    'Content-type': 'application/json'
}


@dataclass
class ElasticSearch(DBC):
    port: int = 9200
    type: str = '_doc'

    _ping_prefix = ''

    def __post_init__(self):
        super().__post_init__()
        self.version = int(self._ping_result.json()['version']['number'][0])
        self.type_url = f'{self.type}/' if self.version <= 6 else ''

    def write(self, index: str, data: Iterator[dict], batch_size=10000, timeout=180):
        self._batch_write(index=index, ndjsons=('{"index":{}}\n' + json.dumps(item) + '\n' for item in data),
                          batch_size=batch_size, timeout=timeout)

    @http_json_res_parse
    def query(self, index, query_body: dict, sort_by_score=False, create_scroll=False, timeout=60):
        if 'sort' not in query_body and not sort_by_score:
            # 不要求按照分数排序搜索会更快一些
            query_body['sort'] = ['_doc']
        url = f'{self.base_url}/{index}/_search{f"?scroll={timeout // 60}m" if create_scroll else ""}'
        return requests.get(url, headers=_HEADERS, data=json.dumps(query_body), timeout=timeout)

    def scroll_query(self, index, query_body: dict, batch_size=1000, timeout=180):
        if 'size' not in query_body:
            query_body['size'] = batch_size
        result = self.query(index, query_body, create_scroll=True, timeout=timeout)
        expect_count = result['hits']['total'] if self.version <= 6 else result['hits']['total']['value']
        scroll_url = f'{self.base_url}/_search/scroll'
        scroll_data = json.dumps({'scroll_id': result['_scroll_id'], 'scroll': f'{timeout // 60}m'})

        while True:
            res = requests.post(
                scroll_url, data=scroll_data,
                headers=_HEADERS, timeout=timeout
            ).json()
            if 'error' in res or not res['hits']['hits']:
                if len(result['hits']['hits']) != expect_count:
                    Logger.warning("查询结果条数与预期不相等，请尝试调大timeout参数或者检查网络")
                return result

            result['took'] += res['took']
            result['hits']['hits'] += (res['hits']['hits'])

    @http_json_res_parse
    def delete(self, index_pattern):
        return requests.delete(f'{self.base_url}/{index_pattern}')

    @http_json_res_parse(is_return=False)
    def _write(self, index, ndjson_data: str, timeout):
        return requests.post(
            # 如果url没有指定index，则调用方在action中指定
            url=f'{self.base_url}/{f"{index}/" if index else "/"}{self.type_url}_bulk',
            data=ndjson_data, headers=_HEADERS, timeout=timeout
        )

    def _batch_write(self, index, ndjsons: Generator[str, None, None], batch_size, timeout):
        self._check_template(index if index.endswith("*") else re.split(r'\W', index)[0] + '*')
        while True:
            items = list(islice(ndjsons, batch_size))
            if not items:  # ES bulk操作body不能为空
                break
            self._write(index=index, ndjson_data=''.join(items), timeout=timeout)

    def _check_template(self, index_pattern):
        url = f'{self.base_url}/_template/{TEMPLATE_NAME}'
        current_templates = requests.get(url).json()
        patterns = current_templates[TEMPLATE_NAME]['index_patterns'] if current_templates else []
        if index_pattern not in patterns:
            current_templates = {
                "index_patterns": patterns + [index_pattern],
                "settings": {
                    "number_of_replicas": 0,
                    "number_of_shards": 3
                },
                "mappings": {
                    "dynamic_date_formats": [
                        "yyyy-MM-dd HH:mm:ss.SSSZ||epoch_millis",
                        "yyyy-MM-dd HH:mm:ss.SSSSSSZ||epoch_millis",
                        "yyyy-MM-dd HH:mm:ss.SSSSSSSSSZ||epoch_millis",
                        "yyyy-MM-dd HH:mm:ssZ||epoch_millis"
                    ],
                    "dynamic_templates": [
                        {
                            "strings_as_keywords": {
                                "match_mapping_type": "string",
                                "mapping": {
                                    "type": "keyword"
                                }
                            }
                        }
                    ]
                }
            }
            if self.version <= 6:
                current_templates['mappings'] = {self.type: current_templates['mappings']}
            self.put_templates(current_templates, TEMPLATE_NAME)

    @http_json_res_parse
    def put_templates(self, templates: dict, template_name):
        url = f'{self.base_url}/_template/{template_name}'
        return requests.put(url, headers=_HEADERS, data=json.dumps(templates))

    def _patch_pandas(self):
        import pandas as pd
        import numpy as np
        from pandas.core.dtypes.dtypes import DatetimeTZDtype

        def to_es(inner_self: pd.DataFrame, index=None, index_col=None, numeric_detection=False,
                  batch_size=10000, timeout=180, copy=True):
            _self = inner_self.copy() if copy else inner_self
            if _self.empty:
                return

            if index and index_col:
                raise ValueError('index和index_col参数不能同时指定')
            if not (index or index_col):
                raise ValueError('index和index_col参数必须指定其中的一个')
            if index:
                index_col = '__$@index'
                _self[index_col] = index
            _self.index = _self.pop(index_col)

            for col, dtype in zip(_self.columns, _self.dtypes):
                if dtype == np.dtype('<M8[ns]'):
                    _self[col] = _self[col].astype('str') + '+0800'
                elif isinstance(dtype, DatetimeTZDtype):
                    # es无法识别"+08:00"的时区标识
                    _self[col] = _self[col].astype('str').apply(lambda x: ''.join(x.rsplit(':', 1)))
                elif numeric_detection and dtype == np.object_:
                    _self[col] = catch(except_return=_self[col], print_traceback=False)(
                        lambda: _self[col].astype('float')
                    )()
            ndjsons = (
                f"{json.dumps({'index': {'_index': name_tuple[0]}})}\n"
                f"{json.dumps(dict(zip(_self.columns, name_tuple[1:])))}\n"
                for name_tuple in _self.itertuples()
            )
            self._batch_write(index=_self.index[0], ndjsons=ndjsons, batch_size=batch_size, timeout=timeout)

        def read_es(index, query_body: dict, batch_size=1000, timeout=180):
            return pd.DataFrame([
                hit['_source']
                for hit in self.scroll_query(index, query_body, batch_size, timeout)['hits']['hits']
            ])

        pd.DataFrame.to_es = to_es
        pd.read_es = read_es



import re
from datetime import datetime
from dateutil import tz


def _split_formats(format_str):
    return re.split(r'(%.)', format_str)


class Datetime(datetime):
    _DEFAULT_TZ = tz.gettz('Asia/Shanghai')
    _DEFAULT_FORMAT = '%Y-%m-%d %H:%M:%S.%f %z'
    _FORMATS = _split_formats(_DEFAULT_FORMAT)
    _SEP_PATTERN = re.compile(r'[-:.Tt ]')

    def __new__(cls, *args, **kwargs):
        kwargs.update(tzinfo=cls._DEFAULT_TZ)
        # args 只保留前7位时间日期参数
        return super().__new__(cls, *args[:7], **kwargs)

    @classmethod
    def fromtimestamp(cls, ts) -> 'Datetime':
        len_ts = len(str(int(ts)))
        if (len_ts - 10) % 3 != 0:
            raise ValueError(f'KPI数据时间戳格式不符合规范，必须是s、ms、us、ns中的一种，当前（{ts}）')
        else:
            ts = ts // 10 ** (len_ts - 10)
        return super().fromtimestamp(ts)

    def to_str(self, start=0, end=6) -> str:
        if not 0 <= start < end <= 8:
            raise ValueError(f'输出的位数参数必须在[0,{len(self._FORMATS) // 2})之间')
        return self.strftime(''.join(self._FORMATS[start * 2 + 1:end * 2]))

    @property
    def str(self):
        return self.to_str()

    @classmethod
    def from_str(cls, datetime_str: str, seps=None):
        if seps is None:
            values = cls._SEP_PATTERN.split(datetime_str)
        else:
            values = re.split(f'[{"".join(seps)}]', datetime_str)
        return Datetime(*map(lambda x: int(x.strip()), values))

    @classmethod
    def from_datetime(cls, dt: datetime):
        return Datetime(*[
            getattr(dt, attr)
            for attr in ['year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond']
        ])

    def __add__(self, other):
        dt = super().__add__(other)
        return Datetime.from_datetime(dt)


def set_default_tz(tz_id):
    Datetime._DEFAULT_TZ = tz.gettz(tz_id)


def set_default_format(format_str):
    Datetime._DEFAULT_FORMAT = format_str
    Datetime._FORMATS = _split_formats(format_str)



import unittest
import pandas as pd
from bools.dbc import ElasticSearch

pd.set_option('display.max_columns', 500)


class ElasticSearchTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        """
        data = pd.read_csv('C:/Users/cb229/Desktop/1231.csv')
        data = data[filter(lambda x: not x.startswith('_'), data.columns)]
        data['datetime'] = pd.to_datetime(data.pop('timestamp'))
        data['date'] = pd.to_datetime(data['date'], utc=True)
        data['date_uct+8'] = data['date'].apply(lambda x: x.tz_convert('Asia/shanghai'))
        data['resp_time'] = data['resp_time'].apply(lambda x: x.replace(',', ''))
        data['index'] = ['test-2020-12-14', 'test-2020-12-13'] * (data.shape[0] // 2) + ['test-2020-12-13']
        self.data = data
        print(data.shape)
        # print(data.head())
        """
        self.es = ElasticSearch(host='http://10.0.80.167', patch_pandas=True)
        # self.es.delete('test*')

    @classmethod
    def test_init(cls):
        ElasticSearch()
        # ElasticSearch(host='http://10.0.80.167', patch_pandas=False)
        # ElasticSearch(user='elastic', password='cb1998827', patch_pandas=True)
        pass

    def test_write(self):
        # self.es.delete('test*')
        # self.es.write('test', [{'a': 1, 'b': 2}] * 20000)
        pass

    def test_es_write(self):
        # self.es.delete('test')
        # pd.DataFrame({'a': [1, 2] * 20000, 'b': range(40000)}).to_es(index='test')
        # self.data.to_es(numeric_detection=True, index='test')
        pass

    def test_query(self):
        # self.es.delete('test*')
        # pd.DataFrame({'a': [1, 2] * 20000, 'b': range(40000)}).to_es(index='test')
        # assert len(self.es.query('test', {
        #     "size": 2000,
        #     "query": {
        #         "terms": {
        #             "a": [1]
        #         }
        #     }
        # })['hits']['hits']) == 2000
        len_ = len(self.es.scroll_query('test', {
            "query": {
                "terms": {
                    "a": [1]
                }
            }
        })['hits']['hits'])
        print(len_)
        assert len_ == 20000

    def test_read_es(self):
        print(pd.read_es('test', {
            "query": {
                "terms": {
                    "a": [1]
                }
            }
        }))


if __name__ == '__main__':
    unittest.main()



import json
import requests
from json.decoder import JSONDecodeError
from functools import wraps
from dataclasses import dataclass
from abc import abstractmethod, ABC


@dataclass
class DBC(ABC):
    host: str = '127.0.0.1'
    port: int = None
    user: str = ''
    password: str = ''
    patch_pandas: bool = False

    version = None
    _ping_prefix = None
    _ping_result = None

    def __post_init__(self):
        self.host = self.host.lstrip('https://').strip()
        self.base_url = f'http://{f"{self.user}:{self.password}@" if self.user else ""}{self.host}:{self.port}'
        if self._ping_prefix is not None:
            ping = requests.get(f'{self.base_url}{self._ping_prefix}')
            if ping.status_code != 200:
                raise ConnectionError(f'无法连接到{self.__class__.__name__}服务器，请检查配置是否正确\n\t{ping.text}')
            self._ping_result = ping

        if self.patch_pandas:
            self._patch_pandas()

    @abstractmethod
    def query(self, *args, **kwargs):
        pass

    @abstractmethod
    def write(self, *args, **kwargs):
        pass

    @abstractmethod
    def _patch_pandas(self):
        pass


def http_json_res_parse(_func=None, *, is_return=True):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            res = func(*args, **kwargs)
            if res.status_code >= 300:
                raise ConnectionError(f'操作执行失败，错误码：[{res.status_code}]\n{res.text}')
            else:
                if is_return:
                    try:
                        return json.loads(res.text)
                    except JSONDecodeError:
                        raise ValueError(f'返回对象不是JSON字符串\n{res.text}')

        return wrapper

    return decorator(_func) if _func else decorator



import unittest
import sys
import os
import pandas as pd

base_dir = f'{os.path.dirname(os.path.abspath(""))}/bools'
sys.path.append(base_dir)

from bools.dbc import InfluxDB
from bools.datetime import Datetime, Timedelta


class InfluxDBTest(unittest.TestCase):
    influxdb = InfluxDB(database='bowaer', user='bowaer', password='cb1998827', patch_pandas=True)
    # influxdb.drop_measurement('test')

    def test_init(self):
        print(self.influxdb.version)

    def test_query(self):
        # print(self.influxdb.query('select * from edges'))
        self.influxdb.query('select * from edges', batch_size=50)
        self.influxdb.query('select * from edges', database='abc')
        self.influxdb.query('select * from edges', database='abc')
        # print(self.influxdb.query('abc * from edges', database='abc'))  # test error sql

    @staticmethod
    def test_pandas_read():
        df = pd.read_influxdb('select * from "test","abc"')
        # assert df.shape[0] == 100
        print(df.head())
        print(df.shape)

    def test_action(self):
        self.influxdb.create_database('bowaer')
        self.influxdb.create_database('test')
        self.influxdb.create_database('def')
        self.influxdb.drop_database('def')
        self.influxdb.drop_database('test')
        self.influxdb.drop_measurement('edges')

    def test_write(self):
        # self.influxdb.drop_measurement('test')
        # self.influxdb.write([
        #     f'test,src=test count={i},succ_count={i} {int(Datetime.now().timestamp()+i*60)*1000000000}'
        #     for i in range(50)
        # ], precision='', batch_size=13)
        pass

    def test_pandas_write(self):
        # self.influxdb.drop_measurement('test')
        # pd.DataFrame({
        #     'time': [Datetime.now().timestamp() + i for i in range(100000)],
        #     # 'measurement': ['edges'] * 10,
        #     'src': ['bowaer'] * 100000,
        #     'count': range(100000),
        #     'succ_count': range(100000)
        # }).to_influxdb(measurement='test', time_col='time', tag_cols=['src'])
        pass

    def test_time(self):
        from bools.functools import timeit
        from influxdb import DataFrameClient
        data = pd.DataFrame({
            'time': [(Datetime.now() + Timedelta(minutes=i)).str for i in range(100000)],
            # 'measurement': ['edges'] * 10,
            'src': ['bowaer'] * 100000,
            'count': range(100000),
            'succ_count': range(100000)
        })
        data.index = data.pop('time')
        # influx = DataFrameClient(host='10.0.80.167', port=7076, database='test', username='bowaer',
        #                          password='cb1998827')
        #
        # timeit(lambda: influx.write_points(
        #     data, 'test', tag_columns=['src'], protocol='line'
        # ))()
        timeit(lambda: data.to_influxdb(measurement='test', tag_cols=['src'], batch_size=10000))()
        # pass


if __name__ == '__main__':
    unittest.main()



import requests
from dataclasses import dataclass
from typing import Generator, Iterator, Union
from itertools import islice
from collections import namedtuple

from .dbc import DBC, http_json_res_parse

_CREATE, _DROP = 'CREATE', 'DROP'
_DATABASE, _MEASUREMENT = 'database', 'measurement'


@dataclass
class InfluxDB(DBC):
    port: int = 8086
    database: str = None

    _ping_prefix = '/ping?verbose=true'

    def __post_init__(self):
        super().__post_init__()
        self.query_url = f'{self.base_url}/query'
        self.write_url = f'{self.base_url}/write'
        self.version = int(self._ping_result.json()['version'][0])

    @http_json_res_parse
    def query(self, influxql: str, database: str = None, batch_size=10000, timeout=180):
        database = self._check_database(database)
        query_url = f'{self.query_url}?db={database}&pretty=false&chunked={batch_size}&q={influxql}'
        return requests.get(query_url, timeout=timeout)

    def write(self, points: Union[Iterator, Generator], database: str = None, precision='n',
              batch_size=10000, timeout=180):
        database = self._check_database(database)
        points = (point for point in points)
        while True:
            items = list(islice(points, batch_size))
            if not items:
                break
            self._write(points=items, database=database, precision=precision, timeout=timeout)

    @http_json_res_parse(is_return=False)
    def _write(self, points: list, database, precision, timeout):
        write_url = f'{self.write_url}?db={database}&precision={precision}'
        return requests.post(write_url, data='\n'.join(points), timeout=timeout)

    def drop_measurement(self, measurement: str, database: str = None):
        self.action(f'{_DROP} {_MEASUREMENT} "{measurement}"', self._check_database(database))

    def drop_database(self, database: str = None):
        self.action(f'{_DROP} {_DATABASE} "{database}"', self._check_database(database))

    def create_database(self, database):
        self.action(f'{_CREATE} {_DATABASE} "{database}"')

    @http_json_res_parse(is_return=False)
    def action(self, influxql, database=''):
        return requests.post(f'{self.query_url}?db={database}&q={influxql}')

    def _patch_pandas(self):
        import pandas as pd
        import numpy as np
        from pandas.core.dtypes.dtypes import DatetimeTZDtype

        def read_influxdb(influxql: str, database: str = None, batch_size=10000, timeout=180, tz_id='Asia/Shanghai'):
            res = self.query(
                influxql=influxql, database=database, batch_size=batch_size, timeout=timeout
            )['results'][0]
            result = pd.DataFrame()

            for block in res.get('series', []):
                df = pd.DataFrame(block['values'], columns=block['columns'])
                df.index = pd.to_datetime(df.pop('time'), utc=True).apply(lambda x: x.tz_convert(tz_id))
                result = pd.concat([result, df], axis=1)
            return result

        def to_influxdb(inner_self: pd.DataFrame, tag_cols, time_col='index',
                        measurement=None, measurement_col=None, database: str = None,
                        batch_size=10000, timeout=180, copy=True):
            _self = inner_self.copy() if copy else inner_self
            if _self.empty:
                return
            if time_col != 'index':
                _self.index = _self.pop(time_col)

            if measurement and measurement_col:
                raise ValueError('measurement和measurement_col参数不能同时指定')
            if not (measurement or measurement_col):
                raise ValueError('measurement和measurement_col参数必须指定其中的一个')
            if measurement:
                measurement_col = f'{_MEASUREMENT}_bowaer2021'
                _self[measurement_col] = measurement

            # iter tuples迭代会修改非命名列名. e.g #_test => _1
            name_transfers = dict(zip(_self.columns, namedtuple('influxdb', _self.columns, rename=True)._fields))

            # hash顺序的字段写入更快
            tag_cols = {name_transfers[col] for col in (tag_cols or [])}
            field_cols = {name_transfers[col] for col in (set(_self.columns) - tag_cols - {measurement_col})}

            try:
                time_dtype = _self.index.dtype
                if time_dtype in {np.dtype(t) for t in ['int32', 'int64', 'float32', 'float64']}:
                    """
                    python3.9  pandas1.2.4  numpy 1.19.3下出现此报错，升级numpy可解决
                    >>> time_dtype in [np.dtype(t) for t in ['int32', 'int64', 'float32', 'float64']]
                        Traceback (most recent call last):
                          File "<stdin>", line 1, in <module>
                        TypeError: Cannot interpret 'datetime64[ns, tzfile('PRC')]' as a data type
                    """
                    _self.index = _self.index.astype('int')
                    power = 19 - len(str(_self.index[0]))
                    if power < 0 or power > 9:
                        raise ValueError(f'时间列：[{time_col}]格式不支持，时间戳支持ns,us,ms,s。请确认是否指定有效时间列')

                    _self.index = _self.index * 10 ** power
                elif time_dtype == np.dtype('O'):
                    _self.index = pd.to_datetime(_self.index)
                    time_dtype = _self.index.dtype

                if time_dtype == np.dtype('<M8[ns]'):
                    # 无时区数据当作UTC+8处理
                    _self.index = _self.index.astype('int') - 8 * 3600 * int(1e9)
                elif isinstance(time_dtype, DatetimeTZDtype):
                    _self.index = _self.index.astype('int')
            except Exception:
                raise ValueError(f'时间列：[{time_col}]格式不支持，当前支持时间戳(ns,us,ms,s), 时间字符串及date列类型')

            points = (
                f'{getattr(name_tuple, measurement_col)}'
                f'{"".join(f",{tag}={getattr(name_tuple, tag)}" for tag in tag_cols)}'
                f' {",".join(f"{field}={getattr(name_tuple, field)}" for field in field_cols)} {name_tuple[0]}'
                for name_tuple in _self.itertuples()
            )
            self.write(points=points, database=database, batch_size=batch_size, timeout=timeout)

        pd.read_influxdb = read_influxdb
        pd.DataFrame.to_influxdb = to_influxdb

    def _check_database(self, database):
        if not (self.database or database):
            raise ValueError('请指定database（初始化传入或函数入参）')
        return database or self.database
